% This file is released under the terms of the GNU General Public License,
% version 3. See http://www.gnu.org/licenses/gpl.html

function [exp1,exp2,exp_full] = Scaling_PSD(plot_path,data,fs,hamming_exponent,PSD_LowFrequencyFit1,PSD_HighFrequencyFit1,PSD_LowFrequencyFit2,PSD_HighFrequencyFit2,PSD_plot)
%
% modified k.linkenkaer@nih.knaw.nl, 031212.
% modified poil@get2net.dk, 07-december-2007, (added improved figure
% handling
%
%
%******************************************************************************************************************
% Purpose...
%
% To compute and plot the power spectrum (PSD) of the input vector 'F' (e.g., abs(W)).
% Fit a power-law function to obtain the power-law scaling exponent.
% Plot the PSD in double logarithmic coordinates.
%
%******************************************************************************************************************
% supporting scripts:
%
% logbinning.m
%
%******************************************************************************************************************
% input parameters...
%
% data	  		: input (column) vector.
% fs			: sampling frequency.
% PSD_LowFrequencyFit	: smallest time scale (window size) to include in power-law fit (in units of seconds!).
% PSD_HighFrequencyFit	: largest time scale (window size) to include in power-law fit (in units of seconds!).
% PSD_plot		: should either be an axeshandle or an integer > 0
%
%******************************************************************************************************************
% output parameters...
%
% PSD_logx		: the logarithmically binned frequency vector (x-axis), BUT THE LOG10 HAS NOT BEEN TAKEN!
% PSD_logy		: the PSD, BUT THE LOG10 HAS NOT BEEN TAKEN!
% PSD_exp		: the PSD power-law exponent.
% PSD_conf		: 95% confidence interval for the power-law exponent.
%
%******************************************************************************************************************
% default parameters...

res_logbin = 50;		% resolution of the logarithmic binning. normal = 7 (10 used in HBM)

%******************************************************************************************************************

% [P,S]=psd(data,length(data),fs,hanning(length(data)));

% [P,S]=psd(data,length(data),fs,hanning(length(data)));

[P,S]=pwelch(data,hamming(2^hamming_exponent),[],[], fs);

%**********************************************************************************************************************
% we want equidistant data points on a loglog scale...

[PSD_logx,PSD_logy] = Logbinning_v2(S,P,res_logbin);

%**********************************************************************************************************************
% now we are ready for loglog regression.

PSD_LowFrequencyFit_log1 = min(find(PSD_logx >= PSD_LowFrequencyFit1));
PSD_HighFrequencyFit_log1 = max(find(PSD_logx <= PSD_HighFrequencyFit1));

SX1=[ones(PSD_HighFrequencyFit_log1-PSD_LowFrequencyFit_log1 + 1,1) log10(PSD_logx(PSD_LowFrequencyFit_log1:PSD_HighFrequencyFit_log1))'];
PY1=log10(PSD_logy(PSD_LowFrequencyFit_log1:PSD_HighFrequencyFit_log1))';
num_indexes = find(~isnan(PY1) & ~isinf(PY1));
SX1=SX1(num_indexes,1:2);
PY1=PY1(num_indexes);
[PSD_exp1,bint1,r1,rint1,stats1]=regress(PY1,SX1);				% psd_exp: the psd_exp of the lsline.
% r: the res_psdidual, the difference between observed and predicted values.
PSD_exp1 = PSD_exp1(2,1);									% stats: contains the R^2 value of the fit (+ see manual).
PSD_conf1=((bint1(2,2))-(bint1(2,1)))/2;					% compute +- 95% confidence intervals

PSD_LowFrequencyFit_log2 = min(find(PSD_logx >= PSD_LowFrequencyFit2));
PSD_HighFrequencyFit_log2 = max(find(PSD_logx <= PSD_HighFrequencyFit2));

SX2=[ones(PSD_HighFrequencyFit_log2-PSD_LowFrequencyFit_log2 + 1,1) log10(PSD_logx(PSD_LowFrequencyFit_log2:PSD_HighFrequencyFit_log2))'];
PY2=log10(PSD_logy(PSD_LowFrequencyFit_log2:PSD_HighFrequencyFit_log2))';
num_indexes = find(~isnan(PY2) & ~isinf(PY2));
SX2=SX2(num_indexes,1:2);
PY2=PY2(num_indexes);
[PSD_exp2,bint2,r2,rint2,stats2]=regress(PY2,SX2);				% psd_exp: the psd_exp of the lsline.
% r: the res_psdidual, the difference between observed and predicted values.
PSD_exp2 = PSD_exp2(2,1);									% stats: contains the R^2 value of the fit (+ see manual).
PSD_conf2=((bint2(2,2))-(bint2(2,1)))/2;					% compute +- 95% confidence intervals

SX_full=[ones(PSD_HighFrequencyFit_log2-PSD_LowFrequencyFit_log1 + 1,1) log10(PSD_logx(PSD_LowFrequencyFit_log1:PSD_HighFrequencyFit_log2))'];
PY_full=log10(PSD_logy(PSD_LowFrequencyFit_log1:PSD_HighFrequencyFit_log2))';
num_indexes = find(~isnan(PY_full) & ~isinf(PY_full));
SX_full=SX_full(num_indexes,1:2);
PY_full=PY_full(num_indexes);
[PSD_exp_full,bint_full,r_full,rint_full,stats_full]=regress(PY_full,SX_full);				% psd_exp: the psd_exp of the lsline.
% r: the res_psdidual, the difference between observed and predicted values.
PSD_exp_full = PSD_exp_full(2,1);									% stats: contains the R^2 value of the fit (+ see manual).
PSD_conf_full=((bint_full(2,2))-(bint_full(2,1)))/2;					% compute +- 95% confidence intervals

%*************************************************************************************************************************
% Plotting the log-binned power spectrum!

if (PSD_plot~=0)
   %plot output
%    if ~ishandle(PSD_plot)		%see if any figure handle is set
%        figure(PSD_plot)
%        PSD_plot = axes;
%    end
%    axes(PSD_plot)
    
    fig=figure('name','PSD log10log10');
    fig.PaperPositionMode = 'auto';
    fig.Position=[100, 100, 1000, 800];
    fig.Visible='off';
    hold on;

    x=log10(PSD_logx(PSD_LowFrequencyFit_log1:PSD_HighFrequencyFit_log1));
    y=log10(PSD_logy(PSD_LowFrequencyFit_log1:PSD_HighFrequencyFit_log1));
    num_indexes = find(~isnan(y) & ~isinf(y));
    x=x(num_indexes);
    y=y(num_indexes);
    plot(x,y,'bo') 				% plotting the NOT normalized power spectrum.
    h=line( x, polyval(polyfit(x,y,1),x));
    set(h(1),'color','b');
    
    x=log10(PSD_logx(PSD_LowFrequencyFit_log2:PSD_HighFrequencyFit_log2));
    y=log10(PSD_logy(PSD_LowFrequencyFit_log2:PSD_HighFrequencyFit_log2));
    plot(x,y,'ro') 				% plotting the NOT normalized power spectrum.
    h=line( x, polyval(polyfit(x,y,1),x));
    set(h(1),'color','r');
    
    plot(log10(PSD_logx(2:length(PSD_logx))),log10(PSD_logy(2:length(PSD_logy))),'k.')		% we fit to the specified interval but plot all data points.
    %axis([log10(min(PSD_logx)) log10(max(PSD_logx)) log10(min(PSD_logx))-1 log10(max(PSD_logx))+1])
    
    axis([log10(min(PSD_logx)) log10(max(PSD_logx)) -5 5])
    
    disp([log10(min(PSD_logx))-1 log10(max(PSD_logx))+1]);
    grid on
    zoom on
    
    title(['PSD-exp(',num2str(PSD_LowFrequencyFit1),',',num2str(PSD_HighFrequencyFit1),')=',num2str(PSD_exp1,3) ,'; PSD-exp(',num2str(PSD_LowFrequencyFit2),',',num2str(PSD_HighFrequencyFit2),')=',num2str(PSD_exp2,3),'; PSD-exp(',num2str(PSD_LowFrequencyFit1),',',num2str(PSD_HighFrequencyFit2),')=',num2str(PSD_exp_full,3) ])
    
    saveas(gcf,strcat(plot_path,'PSD_fit.png'));
    %print(strcat(path,'PSD_fit.png'),'-dpng');
    close(gcf);
    
    exp1=PSD_exp1;
    exp2=PSD_exp2;
    exp_full=PSD_exp_full;
   
end