#     Originally created by Arthur-Ervin Avramiea (2018), a.e.avramiea@vu.nl

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import numpy as np
import cros_model.params as p
from scipy.optimize  import minimize

def get_sum_circle(A,grid,locRange,conP,totalCells):
    sumtoo = 0

    for i in xrange(-locRange,locRange+1):
        for j in xrange(-locRange,locRange+1):
            multi = grid[i+locRange][j+locRange]*A
            #actually this doesn't make any difference because value at this position should be 0 (outside range).
            if math.pow( i*i+j*j,0.5 )>locRange:
                continue
            if multi>1:
                sumtoo+=1
            else:
                sumtoo+=multi

    return abs(sumtoo - conP*totalCells)

def get_sum_square(A,grid,locRange,conP,totalCells):
    sumtoo = 0

    for i in xrange(-locRange,locRange+1):
        for j in xrange(-locRange,locRange+1):
            multi = grid[i+locRange][j+locRange]*A
            #actually this doesn't make any difference because value at this position should be 0 (outside range).
            if multi>1:
                sumtoo+=1
            else:
                sumtoo+=multi

    return abs(sumtoo - conP*totalCells)


def calc_A_exponential_circle(conP,locRange):
    grid = np.zeros((locRange*2+1,locRange*2+1))
    totalCells = 0

    for i in xrange(-locRange,locRange+1):
        for j in xrange(-locRange,locRange+1):
            grid[i+locRange][j+locRange]=math.exp(-math.pow(i*i+j*j,0.5))
            if math.pow( i*i+j*j,0.5 )>locRange or (i==0 and j==0):
                grid[i+locRange][j+locRange]=0
            else:
                totalCells+=1

    #total cells here are the cells within the circle - not to be confused with cells within whole square
    # print "total cells exponential",totalCells

    res = minimize(get_sum_circle,0.1,args=(grid,locRange,conP,totalCells),method='nelder-mead',options={'xtol':1,'disp':False})
    return res.x[0]


def calc_A_exponential_square(conP,locRange):
    grid = np.zeros((locRange*2+1,locRange*2+1))
    totalCells = 0

    for i in xrange(-locRange,locRange+1):
        for j in xrange(-locRange,locRange+1):
            grid[i+locRange][j+locRange]=math.exp(-math.pow(i*i+j*j,0.5))
            if grid[i+locRange][j+locRange]>locRange or (i==0 and j==0):
                grid[i+locRange][j+locRange]=0
            else:
                totalCells+=1

    #total cells here are the cells within the circle - not to be confused with cells within whole square
    # print "total cells exponential",totalCells

    res = minimize(get_sum_square,0.1,args=(grid,locRange,conP,totalCells),method='nelder-mead',options={'xtol':1,'disp':False})
    return res.x[0]

def get_lineardist_multiplier(ratio,crt_dist,max_dist):
    return 1.0 + (crt_dist/float(max_dist))*(ratio-1.0)

def calc_A_linear_circle(conP,locRange,ratio_smallest_highest):
    assert ratio_smallest_highest <=1 and ratio_smallest_highest >=0
    grid = np.zeros((locRange*2+1,locRange*2+1))
    totalCells = 0

    for i in xrange(-locRange,locRange+1):
        for j in xrange(-locRange,locRange+1):
            grid[i+locRange][j+locRange]=get_lineardist_multiplier(ratio_smallest_highest,math.pow(i*i+j*j,0.5),locRange)
            if math.pow( i*i+j*j,0.5 )>locRange or (i==0 and j==0):
                grid[i+locRange][j+locRange]=0
            else:
                totalCells+=1

    #total cells here are the cells within the circle - not to be confused with cells within whole square
    # print "total cells linear",totalCells

    res = minimize(get_sum_circle,0.1,args=(grid,locRange,conP,totalCells),method='nelder-mead',options={'xtol':1,'disp':False})
    return res.x[0]

#get neuron coordinates based on index
def get_coordinates(i,grid_size):
    row=int(math.ceil(i/grid_size))
    col = i%grid_size
    return [row,col]

def get_neuron_idx(row,col,grid_size):
    neuron_idx = row*grid_size+col
    return neuron_idx

def get_rewired_synapses(synapses,indexes_from,indexes_to,num_rewirings):
    edge_indexes_to_rewire = np.random.choice(np.arange(np.shape(synapses)[0]),num_rewirings,replace=False)
    
    #sample twice the number needed just to make sure there are enough unique
    choice_rewire_from = np.random.choice(indexes_from,num_rewirings*2,replace=True)
    choice_rewire_to = np.random.choice(indexes_to,num_rewirings*2,replace=True)
    
    #get unique choices
    all_choices = np.column_stack((choice_rewire_from,choice_rewire_to))
    b = np.ascontiguousarray(all_choices).view(np.dtype((np.void, all_choices.dtype.itemsize * all_choices.shape[1])))
    _, idx = np.unique(b, return_index=True)
    unique_a = all_choices[idx]
    
    #filter out self-connnections
    self_connections = unique_a[:,0]==unique_a[:,1]
    if len(self_connections)>0:
        unique_a = unique_a[~self_connections]
    #     
    for i in xrange(num_rewirings):
        edge_idx = edge_indexes_to_rewire[i]
        synapses[edge_idx][0]=unique_a[i][0]
        synapses[edge_idx][1]=unique_a[i][1]
    
    return synapses    
    
def get_network(num_neurons,types,parameters):
    excitatory_indexes = np.where(types==1)[0]
    inhibitory_indexes = np.where(types==0)[0]

    if parameters[p.NETWORK_TYPE]==p.NETWORK_TYPE_CIRCLE_NEIGHBORHOOD_EXPONENTIAL_DISTANCE:
        [ee_from_to,ei_from_to,ie_from_to,ii_from_to]=get_network_circle_neighborhood_exponential_circle(num_neurons,types,parameters)
    elif parameters[p.NETWORK_TYPE]==p.NETWORK_TYPE_SQUARE_NEIGHBORHOOD_EXPONENTIAL_DISTANCE:
        [ee_from_to,ei_from_to,ie_from_to,ii_from_to]=get_network_square_neighborhood_exponential(num_neurons,types,parameters)
    elif parameters[p.NETWORK_TYPE]==p.NETWORK_TYPE_CIRCLE_NEIGHBORHOOD_EXPONENTIAL_LINEAR_DISTANCE:
        [ee_from_to,ei_from_to,ie_from_to,ii_from_to]=get_network_circle_neighborhood_explinear(num_neurons,types,parameters)
    elif parameters[p.NETWORK_TYPE]==p.NETWORK_TYPE_RANDOM_CONNECTIVITY:
        [ee_from_to,ei_from_to,ie_from_to,ii_from_to]=get_network_random_connectivity(num_neurons,types,parameters)
    
    if parameters[p.RANDOM_REWIRING_PERCENTAGE]>0:
        num_rewirings_ee = int(round(parameters[p.RANDOM_REWIRING_PERCENTAGE] * np.shape(ee_from_to)[0]))
        num_rewirings_ei = int(round(parameters[p.RANDOM_REWIRING_PERCENTAGE] * np.shape(ei_from_to)[0]))
        num_rewirings_ie = int(round(parameters[p.RANDOM_REWIRING_PERCENTAGE] * np.shape(ie_from_to)[0]))
        num_rewirings_ii = int(round(parameters[p.RANDOM_REWIRING_PERCENTAGE] * np.shape(ii_from_to)[0]))
        
        ee_from_to = get_rewired_synapses(ee_from_to, excitatory_indexes, excitatory_indexes, num_rewirings_ee)
        ei_from_to = get_rewired_synapses(ei_from_to, excitatory_indexes, inhibitory_indexes, num_rewirings_ei)
        ie_from_to = get_rewired_synapses(ie_from_to, inhibitory_indexes, excitatory_indexes, num_rewirings_ie)
        ii_from_to = get_rewired_synapses(ii_from_to, inhibitory_indexes, inhibitory_indexes, num_rewirings_ii)
    
    return [ee_from_to,ei_from_to,ie_from_to,ii_from_to]

def cartesian(arrays, out=None):
    """
    Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.

    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.

    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])

    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n / arrays[0].size
    out[:,0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m,1:])
        for j in xrange(1, arrays[0].size):
            out[j*m:(j+1)*m,1:] = out[0:m,1:]
    return out

def get_neighbor_neurons_exp(all_neighborhood_distances,all_neighborhood_indexes,outgoing_indexes_neighborhood,multiplier):
    euclidean_distance = np.sqrt(np.power(all_neighborhood_distances[outgoing_indexes_neighborhood,0],2) + np.power(all_neighborhood_distances[outgoing_indexes_neighborhood,1],2))
    random_values = np.random.rand(len(outgoing_indexes_neighborhood[0]))
    selected_outgoing_indexes = np.where(random_values<multiplier*np.exp(- euclidean_distance[0,:]))

    if(len(selected_outgoing_indexes[0])>0):
        selected_outgoing_actual_neuron_indexes = (all_neighborhood_indexes[outgoing_indexes_neighborhood])[selected_outgoing_indexes]
    else:
        selected_outgoing_actual_neuron_indexes=[]

    return selected_outgoing_actual_neuron_indexes

def get_neighbor_neurons_linear(all_neighborhood_distances,all_neighborhood_indexes,outgoing_indexes_neighborhood,multiplier,localRange,ratio_lowest_highest):
    euclidean_distance = np.sqrt(np.power(all_neighborhood_distances[outgoing_indexes_neighborhood,0],2) + np.power(all_neighborhood_distances[outgoing_indexes_neighborhood,1],2))
    random_values = np.random.rand(len(outgoing_indexes_neighborhood[0]))
    selected_outgoing_indexes = np.where(random_values<multiplier * ( 1.0+ (euclidean_distance[0,:]/float(localRange))*(ratio_lowest_highest-1.0) ) )

    # print "outgoing indexes neighborhood",outgoing_indexes_neighborhood
    # print "params",multiplier,localRange,ratio_lowest_highest
    # print "euclidean distance",np.shape(euclidean_distance)
    # print "random values",np.shape(random_values)
    # print "get neighbors",np.shape(outgoing_indexes_neighborhood), np.shape(selected_outgoing_indexes),np.shape(all_neighborhood_indexes)
    # print selected_outgoing_indexes

    if(len(selected_outgoing_indexes[0])>0):
        selected_outgoing_actual_neuron_indexes = (all_neighborhood_indexes[outgoing_indexes_neighborhood])[selected_outgoing_indexes]
    else:
        selected_outgoing_actual_neuron_indexes=[]

    return selected_outgoing_actual_neuron_indexes

def get_network_circle_neighborhood_explinear(num_neurons,types,parameters):
    network_size=int(parameters[p.NETWORK_WIDTH])
    neuron_range_exc=int(parameters[p.LOCAL_RANGE_EX])
    neuron_range_inh=int(parameters[p.LOCAL_RANGE_IN])
    connectivity_exc = parameters[p.EXCITATORY_CONNECTIVITY]
    connectivity_inh = parameters[p.INHIBITORY_CONNECTIVITY]
    ratio_smallest_highest = parameters[p.LINEAR_DISTANCE_RATIO]

    ee_from_to = []
    ei_from_to = []
    ie_from_to = []
    ii_from_to = []

    #TODO: make this back linear
    exc_A = calc_A_linear_circle(connectivity_exc,neuron_range_exc,ratio_smallest_highest)
    inh_A = calc_A_exponential_circle(connectivity_inh,neuron_range_inh)

    #TODO: make easier to read code - change variable names etc
    for i in xrange(num_neurons):
        [from_row,from_col]=get_coordinates(i,network_size)
        if types[i]==p.TYPE_EXC:
            crt_range = neuron_range_exc
        else:
            crt_range = neuron_range_inh

        all_neighborhood_distances = cartesian([np.arange(-crt_range,crt_range+1),np.arange(-crt_range,crt_range+1)])
        #get rid of current neuron (no recurrent connections)
        index_of_current_neuron = np.where(np.logical_and(all_neighborhood_distances[:,0]==0 , all_neighborhood_distances[:,1]==0))
        indexes_outside_circle = np.where(np.sqrt(all_neighborhood_distances[:,0]**2 + all_neighborhood_distances[:,1]**2) > crt_range)
        #get rid of indexes outside circle
        get_rid_of_indexes = np.union1d( index_of_current_neuron[0], indexes_outside_circle [0])
        all_neighborhood_distances = all_neighborhood_distances[np.setdiff1d( np.arange(len(all_neighborhood_distances)), get_rid_of_indexes ),:]

        neighborhood_rows = from_row+all_neighborhood_distances[:,0]
        neighborhood_cols = from_col+all_neighborhood_distances[:,1]

        if parameters[p.TORUS_WRAP_AROUND]:
            outside_left = np.where(neighborhood_cols<0)
            neighborhood_cols[outside_left]=network_size+neighborhood_cols[outside_left]
            outside_right = np.where(neighborhood_cols>=network_size)
            neighborhood_cols[outside_right]=neighborhood_cols[outside_right]-network_size
            
            outside_top = np.where(neighborhood_rows<0)
            neighborhood_rows[outside_top]=network_size+neighborhood_rows[outside_top]
            outside_bottom = np.where(neighborhood_rows>=network_size)
            neighborhood_rows[outside_bottom]=neighborhood_rows[outside_bottom]-network_size
            
            within_grid_indexes=np.arange(len(neighborhood_rows))
            legal_neighborhood_distances = all_neighborhood_distances[within_grid_indexes,:]
        else:
            within_grid_indexes=np.logical_and.reduce(np.array( (neighborhood_rows>=0 , neighborhood_cols>=0, neighborhood_rows<network_size, neighborhood_cols<network_size)))
            legal_neighborhood_distances = all_neighborhood_distances[within_grid_indexes,:]
        
        legal_rows = neighborhood_rows[within_grid_indexes]
        legal_cols = neighborhood_cols[within_grid_indexes]
        legal_indexes = legal_rows*network_size+legal_cols

        outgoing_exc_indexes_within_neighborhood = np.where(types[legal_indexes]==p.TYPE_EXC)
        outgoing_inh_indexes_within_neighborhood = np.where(types[legal_indexes]==p.TYPE_INH)

        if types[i]==p.TYPE_EXC:
            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_linear(legal_neighborhood_distances,legal_indexes,outgoing_exc_indexes_within_neighborhood,exc_A,crt_range,ratio_smallest_highest)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ee_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_linear(legal_neighborhood_distances,legal_indexes,outgoing_inh_indexes_within_neighborhood,exc_A,crt_range,ratio_smallest_highest)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ei_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

        elif types[i]==p.TYPE_INH:
            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_exc_indexes_within_neighborhood,inh_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ie_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_inh_indexes_within_neighborhood,inh_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ii_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

    return [ np.array(ee_from_to),np.array(ei_from_to),np.array(ie_from_to),np.array(ii_from_to) ]

def get_network_circle_neighborhood_exponential_circle(num_neurons,types,parameters):
    network_size=int(parameters[p.NETWORK_WIDTH])
    neuron_range_exc=int(parameters[p.LOCAL_RANGE_EX])
    neuron_range_inh=int(parameters[p.LOCAL_RANGE_IN])
    connectivity_exc = parameters[p.EXCITATORY_CONNECTIVITY]
    connectivity_inh = parameters[p.INHIBITORY_CONNECTIVITY]

    ee_from_to = []
    ei_from_to = []
    ie_from_to = []
    ii_from_to = []

    #TODO: make this back linear
    #exc_A = calc_A_linear(connectivity_exc,neuron_range_exc,ratio_smallest_highest)

    exc_A = calc_A_exponential_circle(connectivity_exc,neuron_range_exc)
    inh_A = calc_A_exponential_circle(connectivity_inh,neuron_range_inh)

    #TODO: make easier to read code - change variable names etc
    for i in xrange(num_neurons):
        [from_row,from_col]=get_coordinates(i,network_size)
        if types[i]==p.TYPE_EXC:
            crt_range = neuron_range_exc
        else:
            crt_range = neuron_range_inh

        all_neighborhood_distances = cartesian([np.arange(-crt_range,crt_range+1),np.arange(-crt_range,crt_range+1)])
        #get rid of current neuron (no recurrent connections)
        index_of_current_neuron = np.where(np.logical_and(all_neighborhood_distances[:,0]==0 , all_neighborhood_distances[:,1]==0))
        #get rid of indexes outside circle
        indexes_outside_circle = np.where(np.sqrt(all_neighborhood_distances[:,0]**2 + all_neighborhood_distances[:,1]**2) > crt_range)
        get_rid_of_indexes = np.union1d( index_of_current_neuron[0], indexes_outside_circle [0])
        all_neighborhood_distances = all_neighborhood_distances[np.setdiff1d( np.arange(len(all_neighborhood_distances)), get_rid_of_indexes ),:]

        neighborhood_rows = from_row+all_neighborhood_distances[:,0]
        neighborhood_cols = from_col+all_neighborhood_distances[:,1]

        if parameters[p.TORUS_WRAP_AROUND]:
            outside_left = np.where(neighborhood_cols<0)
            neighborhood_cols[outside_left]=network_size+neighborhood_cols[outside_left]
            outside_right = np.where(neighborhood_cols>=network_size)
            neighborhood_cols[outside_right]=neighborhood_cols[outside_right]-network_size
            
            outside_top = np.where(neighborhood_rows<0)
            neighborhood_rows[outside_top]=network_size+neighborhood_rows[outside_top]
            outside_bottom = np.where(neighborhood_rows>=network_size)
            neighborhood_rows[outside_bottom]=neighborhood_rows[outside_bottom]-network_size
            
            within_grid_indexes=np.arange(len(neighborhood_rows))
            legal_neighborhood_distances = all_neighborhood_distances[within_grid_indexes,:]
        else:
            within_grid_indexes=np.logical_and.reduce(np.array( (neighborhood_rows>=0 , neighborhood_cols>=0, neighborhood_rows<network_size, neighborhood_cols<network_size)))
            legal_neighborhood_distances = all_neighborhood_distances[within_grid_indexes,:]

        legal_rows = neighborhood_rows[within_grid_indexes]
        legal_cols = neighborhood_cols[within_grid_indexes]
        legal_indexes = legal_rows*network_size+legal_cols

        outgoing_exc_indexes_within_neighborhood = np.where(types[legal_indexes]==p.TYPE_EXC)
        outgoing_inh_indexes_within_neighborhood = np.where(types[legal_indexes]==p.TYPE_INH)

        if types[i]==p.TYPE_EXC:
            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_exc_indexes_within_neighborhood,exc_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ee_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_inh_indexes_within_neighborhood,exc_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ei_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

        elif types[i]==p.TYPE_INH:
            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_exc_indexes_within_neighborhood,inh_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ie_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_inh_indexes_within_neighborhood,inh_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ii_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])


    return [ np.array(ee_from_to),np.array(ei_from_to),np.array(ie_from_to),np.array(ii_from_to) ]

def get_network_square_neighborhood_exponential(num_neurons,types,parameters):
    network_size=int(parameters[p.NETWORK_WIDTH])
    neuron_range_exc=int(parameters[p.LOCAL_RANGE_EX])
    neuron_range_inh=int(parameters[p.LOCAL_RANGE_IN])
    connectivity_exc = parameters[p.EXCITATORY_CONNECTIVITY]
    connectivity_inh = parameters[p.INHIBITORY_CONNECTIVITY]

    ee_from_to = []
    ei_from_to = []
    ie_from_to = []
    ii_from_to = []

    #TODO: make this back linear

    exc_A = calc_A_exponential_square(connectivity_exc,neuron_range_exc)
    inh_A = calc_A_exponential_square(connectivity_inh,neuron_range_inh)

    #TODO: make easier to read code - change variable names etc
    for i in xrange(num_neurons):
        [from_row,from_col]=get_coordinates(i,network_size)
        if types[i]==p.TYPE_EXC:
            crt_range = neuron_range_exc
        else:
            crt_range = neuron_range_inh

        all_neighborhood_distances = cartesian([np.arange(-crt_range,crt_range+1),np.arange(-crt_range,crt_range+1)])
        #get rid of current neuron (no recurrent connections)
        index_of_current_neuron = np.where(np.logical_and(all_neighborhood_distances[:,0]==0 , all_neighborhood_distances[:,1]==0))
        get_rid_of_indexes = index_of_current_neuron[0]
        all_neighborhood_distances = all_neighborhood_distances[np.setdiff1d( np.arange(len(all_neighborhood_distances)), get_rid_of_indexes ),:]
        
        neighborhood_rows = from_row+all_neighborhood_distances[:,0]
        neighborhood_cols = from_col+all_neighborhood_distances[:,1]

        if parameters[p.TORUS_WRAP_AROUND]:
            outside_left = np.where(neighborhood_cols<0)
            neighborhood_cols[outside_left]=network_size+neighborhood_cols[outside_left]
            outside_right = np.where(neighborhood_cols>=network_size)
            neighborhood_cols[outside_right]=neighborhood_cols[outside_right]-network_size
            
            outside_top = np.where(neighborhood_rows<0)
            neighborhood_rows[outside_top]=network_size+neighborhood_rows[outside_top]
            outside_bottom = np.where(neighborhood_rows>=network_size)
            neighborhood_rows[outside_bottom]=neighborhood_rows[outside_bottom]-network_size
            
            within_grid_indexes=np.arange(len(neighborhood_rows))
            legal_neighborhood_distances = all_neighborhood_distances[within_grid_indexes,:]
        else:
            within_grid_indexes=np.logical_and.reduce(np.array( (neighborhood_rows>=0 , neighborhood_cols>=0, neighborhood_rows<network_size, neighborhood_cols<network_size)))
            legal_neighborhood_distances = all_neighborhood_distances[within_grid_indexes,:]
            
        legal_rows = neighborhood_rows[within_grid_indexes]
        legal_cols = neighborhood_cols[within_grid_indexes]
        legal_indexes = legal_rows*network_size+legal_cols
        
        outgoing_exc_indexes_within_neighborhood = np.where(types[legal_indexes]==p.TYPE_EXC)
        outgoing_inh_indexes_within_neighborhood = np.where(types[legal_indexes]==p.TYPE_INH)

        if types[i]==p.TYPE_EXC:
            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_exc_indexes_within_neighborhood,exc_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ee_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_inh_indexes_within_neighborhood,exc_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ei_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

        elif types[i]==p.TYPE_INH:
            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_exc_indexes_within_neighborhood,inh_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ie_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])

            selected_outgoing_actual_neuron_indexes=get_neighbor_neurons_exp(legal_neighborhood_distances,legal_indexes,outgoing_inh_indexes_within_neighborhood,inh_A)
            if len(selected_outgoing_actual_neuron_indexes)>0:
                for j in xrange(len(selected_outgoing_actual_neuron_indexes)):
                    ii_from_to.append([i,selected_outgoing_actual_neuron_indexes[j]])


    return [ np.array(ee_from_to),np.array(ei_from_to),np.array(ie_from_to),np.array(ii_from_to) ]

def get_network_random_connectivity(num_neurons,types,parameters):
    ee_from_to = []
    ei_from_to = []
    ie_from_to = []
    ii_from_to = []

    print "building random connectivity"

    for i in xrange(num_neurons):
            connection_probability = np.random.rand(num_neurons)
            indexes_connection_probability = np.where(connection_probability>=(1-parameters[p.CONNECTIVITY_PROBABILITY]))[0]
            for j in list(indexes_connection_probability):
                if j != i:
                    if types[i]==1 and types[j]==1:
                        ee_from_to.append([i,j])
                    elif types[i]==1 and types[j]==0:
                        ei_from_to.append([i,j])
                    elif types[i]==0 and types[j]==1:
                        ie_from_to.append([i,j])
                    elif types[i]==0 and types[j]==0:
                        ii_from_to.append([i,j])

    return [ np.array(ee_from_to),np.array(ei_from_to),np.array(ie_from_to),np.array(ii_from_to) ]