#     Originally created by Arthur-Ervin Avramiea (2018), a.e.avramiea@vu.nl

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

from bitstream import BitStream
import os
import h5py

from bitarray import bitarray
from brian2 import *

import cros_model.params as p
from cros_model.configuration import configuration as conf


def get_output_directory_name(param):
    configuration = conf.get_configuration()
    crt_dir = configuration[conf.DATA_FILE_PATH]
    
    if 'Custom' in param[p.SIMULATION_TYPE]:        
        output_dir_param_keys = param[p.OUTPUT_DIR_PATH_PARAMETERS].split(",")
        output_dir_param_values = []
        for parameter_key in output_dir_param_keys:
            output_dir_param_values.append(param[parameter_key.strip()])
        output_directory_name = param[p.OUTPUT_DIR_PATH_FORMAT] % tuple(output_dir_param_values)
    
    if param[p.SIMULATION_TYPE]=='OriginalNetwork':
        output_directory_name = "OriginalNetwork/%s/OriginalNetwork.CPE%.2f.CI%.2f.Run%d/" % \
                                (param[p.OUTPUT_DIR],param[p.EXCITATORY_CONNECTIVITY],param[p.INHIBITORY_CONNECTIVITY],param[p.RUN_NUMBER])

    if param[p.SIMULATION_TYPE]=='OriginalSimulation':
        output_directory_name = "OriginalSimulation/%s/OriginalSimulationConnect.CPE%.2f.CI%.2f.Run%d/" % \
                                (param[p.OUTPUT_DIR],param[p.EXCITATORY_CONNECTIVITY],param[p.INHIBITORY_CONNECTIVITY],param[p.RUN_NUMBER])
                                
    if param[p.SIMULATION_TYPE]=='NeuromodSimulation':
        output_directory_name = "NeuromodSimulation/%s/NeuromodSimulation.CPE%.2f.CI%.2f.Run%d.WEE%.5f.WEI%.5f.WIE%.5f.WII%.5f.EEG%.2f.EIG%.2f.IEG%.2f.IIG%.2f.S%i/" % \
                                (param[p.OUTPUT_DIR],param[p.EXCITATORY_CONNECTIVITY],param[p.INHIBITORY_CONNECTIVITY],param[p.RUN_NUMBER],\
                                 param[p.W_EX_EX],param[p.W_EX_IN],param[p.W_IN_EX],param[p.W_IN_IN],\
                                 param[p.EE_GAIN],param[p.EI_GAIN],param[p.IE_GAIN],param[p.II_GAIN],param[p.RANDOM_SEED])

    output_directory_name=crt_dir+output_directory_name
    return output_directory_name

def get_bytes_from_stream(stream):
    return bitarray(str(stream))

def get_network_arrays_from_path(path):
    f = h5py.File(path, "r")
    
    types = np.array(f["networks/main_network/types"])
    main_network_from_to=np.array(f["connections/from_main_to_main/indexes"])
    main_network_weights=np.array(f["connections/from_main_to_main/weights"])
    
    ee_indexes = np.where( (types[main_network_from_to[:,0]]==1) & (types[main_network_from_to[:,1]]==1) )[0]
    ei_indexes = np.where( (types[main_network_from_to[:,0]]==1) & (types[main_network_from_to[:,1]]==0) )[0]
    ie_indexes = np.where( (types[main_network_from_to[:,0]]==0) & (types[main_network_from_to[:,1]]==1) )[0]
    ii_indexes = np.where( (types[main_network_from_to[:,0]]==0) & (types[main_network_from_to[:,1]]==0) )[0]
    
    ee_from_to = main_network_from_to[ee_indexes,:]
    wEE = main_network_weights[ee_indexes]
    ei_from_to = main_network_from_to[ei_indexes,:]
    wEI = main_network_weights[ei_indexes]
    ie_from_to = main_network_from_to[ie_indexes,:]
    wIE = main_network_weights[ie_indexes]
    ii_from_to = main_network_from_to[ii_indexes,:]
    wII = main_network_weights[ii_indexes]
    
    # if there is stimulus network
    if "networks/stimulus_network" in f:
        types_stimulus = np.array(f["networks/stimulus_network/types"])
        stimulus_to_main_network_from_to=np.array(f["connections/from_stimulus_to_main/indexes"])
        stimulus_to_main_network_weights=np.array(f["connections/from_stimulus_to_main/weights"])
    else:
        types_stimulus=[]
        stimulus_to_main_network_from_to=np.empty([0,2])
        stimulus_to_main_network_weights=np.empty(0)
        
    f.close()                
    
    return [ types,types_stimulus,\
            ee_from_to,ei_from_to,ie_from_to,ii_from_to,stimulus_to_main_network_from_to,\
            wEE,wEI,wIE,wII,stimulus_to_main_network_weights ]

class Output:
    def __init__(self,params):
        self.params = params
        self.output_dir_path = get_output_directory_name(params)

        print self.output_dir_path

        if not os.path.exists(self.output_dir_path):
            print "creating missing path"
            os.makedirs(self.output_dir_path)

        # self.spikes_outputfile = open(self.output_dir_path+"Time.Series","w")

    def write_spikes_info(self,spike_ms_times,external_input_times):
        f = h5py.File(self.output_dir_path+"SimulationData.h5", "a")
        
        f.create_dataset("data/population_spiking_main_network",spike_ms_times.shape,dtype="f",\
                         data=spike_ms_times, compression="gzip", compression_opts=9)
        
        if len(external_input_times)>0:
            f.create_dataset("data/population_spiking_external_input_network",external_input_times.shape,dtype="f",\
                             data=external_input_times, compression="gzip", compression_opts=9)
        
        f.close()

    def write_network_structure(self,excitatory_indexes_sorted,inhibitory_indexes_sorted,types,\
                                Cee,Cei,Cie,Cii,Cse,stimulus_size):
        
        f = h5py.File(self.output_dir_path+"SimulationData.h5", "a")
        
        f.create_dataset("networks/main_network/types",types.shape,dtype="i",data=types, compression="gzip", compression_opts=9)
        main_network_group = f["networks/main_network"]
        main_network_group.attrs["index_network"]=0
        main_network_group.attrs["type"]="grid_network"
                    
        main_network_from = np.concatenate( (excitatory_indexes_sorted[Cee.i],excitatory_indexes_sorted[Cei.i],\
                                            inhibitory_indexes_sorted[Cie.i],inhibitory_indexes_sorted[Cii.i]) )
        
        main_network_to = np.concatenate( (excitatory_indexes_sorted[Cee.j],inhibitory_indexes_sorted[Cei.j],\
                                           excitatory_indexes_sorted[Cie.j],inhibitory_indexes_sorted[Cii.j]) )
        
        main_network_from_to = np.transpose(np.vstack( (main_network_from,main_network_to) ))        
        main_network_weights = np.concatenate( (Cee.w,Cei.w,Cie.w,Cii.w) )

        f.create_dataset("connections/from_main_to_main/indexes", main_network_from_to.shape, dtype='i', data=main_network_from_to, compression="gzip", compression_opts=9)
        f.create_dataset("connections/from_main_to_main/weights", main_network_weights.shape, dtype='f', data=main_network_weights, compression="gzip", compression_opts=9)
        from_main_to_main_connections = f["connections/from_main_to_main"]
        from_main_to_main_connections.attrs["from_network_index"]=0
        from_main_to_main_connections.attrs["to_network_index"]=0
        
        #extend types with external neurons
        if stimulus_size>0:
            types_stimulus = np.repeat(1,stimulus_size)
            stimulus_network_neuron_types = f.create_dataset("networks/stimulus_network/types",types_stimulus.shape,dtype="i",data=types_stimulus, compression="gzip", compression_opts=9)
            stimulus_network_neuron_types.attrs["index_network"]=1
            stimulus_network_neuron_types.attrs["type"]="random network" 
            
            stimulus_to_main_from = Cse.i
            stimulus_to_main_to = excitatory_indexes_sorted[Cse.j]
            
            stimulus_to_main_from_to = np.transpose( np.vstack((stimulus_to_main_from,stimulus_to_main_to)) )
            stimulus_to_main_weights = Cse.w
            
            f.create_dataset("connections/from_stimulus_to_main/indexes", stimulus_to_main_from_to.shape, dtype='i', data=stimulus_to_main_from_to, compression="gzip", compression_opts=9)
            f.create_dataset("connections/from_stimulus_to_main/weights", stimulus_to_main_weights.shape, dtype='f', data=stimulus_to_main_weights, compression="gzip", compression_opts=9)
            from_stimulus_to_main_connections = f["connections/from_stimulus_to_main"]
            from_stimulus_to_main_connections.attrs["from_network_index"] = 1
            from_stimulus_to_main_connections.attrs["to_network_index"] = 0
        
        f.close()

    def write_params(self):
        with open(self.output_dir_path+"Simulation.Parameters","w") as f:
            for crt_key in sorted(self.params.keys()):
                crt_value = self.params[crt_key]
                f.write(crt_key+"="+str(crt_value)+"\n")

    def write_fitness_value(self,fitness1,fitness2,fitness3,fitness4):
        string_all = ""
        string_all+=str(fitness1)+"\n"
        string_all+=str(fitness2)+"\n"
        string_all+=str(fitness3)+"\n"
        string_all+=str(fitness4)+"\n"
        with open(self.output_dir_path+"ActivityFitness.txt","w") as f:
            f.write(string_all)

    def finalize_writing(self):
        self.spikes_outputfile.close()

    def get_output_dir_path(self):
        return self.output_dir_path

    def output_simulation(self,simulation_variables):
        population_monitor_exc = simulation_variables['population_monitor_exc']
        population_monitor_inh = simulation_variables['population_monitor_inh']
        num_excitatory = simulation_variables['num_excitatory']
        num_inhibitory = simulation_variables['num_inhibitory']
        types = simulation_variables['types']

        Cee = simulation_variables['Cee']
        Cei = simulation_variables['Cei']
        Cie = simulation_variables['Cie']
        Cii = simulation_variables['Cii']
        Cse = simulation_variables['Cse']
        stimulus_size = simulation_variables['stimulus_size']
        external_input_times = simulation_variables['external_input_times']
        
        excitatory_indexes_sorted=simulation_variables['excitatory_indexes_sorted']
        inhibitory_indexes_sorted=simulation_variables['inhibitory_indexes_sorted']
        
        spike_ms_times_exc=np.round(population_monitor_exc.rate/Hz*num_excitatory*0.001)
        spike_ms_times_inh=np.round(population_monitor_inh.rate/Hz*num_inhibitory*0.001)
        
        spike_ms_times_both = np.column_stack((spike_ms_times_exc,spike_ms_times_inh))
        
        #first create new empty simulation data hdf5 file and then append to it
        f = h5py.File(self.output_dir_path+"SimulationData.h5", "w")
        f.close()
        
        self.write_spikes_info(spike_ms_times_both, external_input_times)
        
        with open(self.output_dir_path+"NetworkStatistics.txt","w") as f:
            f.write("eeCount="+str(len(Cee.i))+"\n"+\
                    "eiCount="+str(len(Cei.i))+"\n"+\
                    "ieCount="+str(len(Cie.i))+"\n"+\
                    "iiCount="+str(len(Cii.i)))
        
        with open(self.output_dir_path+"ActivityStatistics.txt","w") as f:
            f.write("meanFiringrateExc="+str(np.mean(population_monitor_exc.rate/Hz))+"\n"+\
                    "meanFiringrateInh="+str(np.mean(population_monitor_inh.rate/Hz))+"\n")
                    
        if self.params[p.OUTPUT_ACTIVITY]:
            s_mon_exc = simulation_variables['s_mon_exc']
            s_mon_inh = simulation_variables['s_mon_inh']
            
            #TODO: if I want networks with higher than 65536 neurons, then turn 16 to 32bits to be able to handle index sizes, limited to 64000 now
            np.save(self.output_dir_path+"output_spikes_neurons_exc.dat",np.array(s_mon_exc.it[0]).astype(np.int16))
            np.save(self.output_dir_path+"output_spikes_spikes_exc.dat",np.array(s_mon_exc.t/ms).astype(np.int32))
            np.save(self.output_dir_path+"output_spikes_neurons_inh.dat",np.array(s_mon_inh.it[0]).astype(np.int16))
            np.save(self.output_dir_path+"output_spikes_spikes_inh.dat",np.array(s_mon_inh.t/ms).astype(np.int32))

        if self.params[p.OUTPUT_NETWORK]:
            self.write_network_structure(excitatory_indexes_sorted,inhibitory_indexes_sorted,types,Cee,Cei,Cie,Cii,Cse,stimulus_size)
        
        if 's_mon_exc_cycles' in simulation_variables.keys():
            s_mon_exc_cycles=simulation_variables['s_mon_exc_cycles']
            cycles_mean = np.mean(s_mon_exc_cycles.maxC)
            cycles_sd = np.std(s_mon_exc_cycles.maxC)
            with open(self.output_dir_path+"CyclesStat.txt","w") as f:
                f.write(str(cycles_mean)+"\n")
                f.write(str(cycles_sd)+"\n")