#     Originally created by Arthur-Ervin Avramiea (2018), a.e.avramiea@vu.nl

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

import shutil
import platform
from time import time, sleep
from brian2 import *
from cros_model.output import output
from cros_model import params as p, network
import random
import os
import numpy_indexed as npi
import cros_model.data_analysis.matlab_analysis as matlab_analysis

#Brian settings
prefs.codegen.target = 'cython'
if "Windows" not in platform.platform():
    prefs.codegen.cpp.extra_compile_args_gcc = ['-w', '-Ofast', '-march=native']

params_file_path = sys.argv[1]
output_file_path = sys.argv[2]

parameters = p.load_params_file(params_file_path,"")
random.seed(parameters[p.RANDOM_SEED])
show_gui=parameters[p.SHOW_SIM_GUI]
# currently not possible to show gui if running in standalone mode
standalone=parameters[p.STANDALONE_MODE] and not show_gui

#initialize output
simulation_output = output.Output(parameters)
simulation_output.write_params()

########################################## CREATE / LOAD NETWORK STRUCTURE

#initialize network structure before the section where I run standalone brian c++ code (code generator cannot handle parts of this section)
num_neurons=int(parameters[p.NUMBER_NEURONS])
types=np.zeros(num_neurons)

num_excitatory = int(round(parameters[p.FRACTION_EXCITATORY]*num_neurons))
num_inhibitory = num_neurons-num_excitatory

if not p.NETWORK_TO_LOAD_PATH in parameters.keys():
    all_indexes = np.array(range(0,num_neurons))
    np.random.shuffle(all_indexes)
    types[all_indexes[0:num_excitatory]]=1

    excitatory_indexes = all_indexes[0:num_excitatory]
    inhibitory_indexes = all_indexes[num_excitatory:num_neurons]
    excitatory_indexes_sorted = np.sort(all_indexes[0:num_excitatory])
    inhibitory_indexes_sorted = np.sort(all_indexes[num_excitatory:num_neurons])

    [ee_from_to,ei_from_to,ie_from_to,ii_from_to]= network.get_network(num_neurons,types,parameters)
        
    wEE = parameters[p.W_EX_EX]
    wEI = parameters[p.W_EX_IN]
    wIE = parameters[p.W_IN_EX]
    wII = parameters[p.W_IN_IN]
    
    if parameters[p.EXTERNAL_INPUT]:
        stimulus_size = parameters[p.STIMULUS_SIZE]
        stimulus_input_from_to = np.column_stack( (np.arange(stimulus_size),np.random.choice(excitatory_indexes,size = stimulus_size,replace = False)) )
        w_external_exc = 0.0085
    else:
        stimulus_size = 0
        stimulus_input_from_to = np.empty([0,2])
        w_external_exc = 0
else:
    #TODO: add in some checks for network size. not needed now
    [types,types_stimulus,ee_from_to,ei_from_to,ie_from_to,ii_from_to,stimulus_input_from_to,wEE,wEI,wIE,wII,w_external_exc]=output.get_network_arrays_from_path(parameters[p.NETWORK_TO_LOAD_PATH])
    
    stimulus_size = len(types_stimulus)
    
    excitatory_indexes = np.where(types==1)[0]
    inhibitory_indexes = np.where(types==0)[0]
    
    excitatory_indexes_sorted = np.sort(excitatory_indexes)
    inhibitory_indexes_sorted = np.sort(inhibitory_indexes)
    
    #TODO: fix
    #this is here because in the current protocol we don't want to recover the input layer from the network
    if parameters[p.EXTERNAL_INPUT]:
        stimulus_size = parameters[p.STIMULUS_SIZE]
        stimulus_input_from_to = np.column_stack( (np.arange(stimulus_size),np.random.choice(excitatory_indexes,size = stimulus_size,replace = False)) )
        w_external_exc = 0.0085
    else:
        stimulus_size = 0
        stimulus_input_from_to = np.empty([0,2])
        w_external_exc = 0

ee_from = ee_from_to[:,0]
ee_to = ee_from_to[:,1]
ei_from = ei_from_to[:,0]
ei_to = ei_from_to[:,1]
ie_from = ie_from_to[:,0]
ie_to = ie_from_to[:,1]
ii_from = ii_from_to[:,0]
ii_to = ii_from_to[:,1]
stimulus_input_from = stimulus_input_from_to[:,0]
stimulus_input_to = stimulus_input_from_to[:,1]

ee_from_idx = npi.indices(excitatory_indexes_sorted, ee_from)
ee_to_idx = npi.indices(excitatory_indexes_sorted, ee_to)
ei_from_idx = npi.indices(excitatory_indexes_sorted, ei_from)
ei_to_idx = npi.indices(inhibitory_indexes_sorted, ei_to)
ie_from_idx = npi.indices(inhibitory_indexes_sorted, ie_from)
ie_to_idx = npi.indices(excitatory_indexes_sorted, ie_to)
ii_from_idx = npi.indices(inhibitory_indexes_sorted, ii_from)
ii_to_idx = npi.indices(inhibitory_indexes_sorted, ii_to)
stimulus_input_from_idx = stimulus_input_from #this doesn't change
stimulus_input_to_idx = npi.indices(excitatory_indexes_sorted,stimulus_input_to)

########################################## NEURON MODEL VARIABLES

background_p_exc=parameters[p.BACKGROUND_P_EXC]
background_p_inh=parameters[p.BACKGROUND_P_INH]
taup_exc=parameters[p.EXC_DECAY]*ms
taup_inh=parameters[p.INH_DECAY]*ms
taui_exc=parameters[p.INPUT_DECAY_EX]*ms
taui_inh=parameters[p.INPUT_DECAY_IN]*ms
P_reset_exc=parameters[p.P_RESET_EXC]
P_reset_inh=parameters[p.P_RESET_INH]
network_size=parameters[p.NETWORK_WIDTH]
num_neurons=parameters[p.NUMBER_NEURONS]

tauI_e=parameters[p.INPUT_DECAY_EX]*ms
tauI_i=parameters[p.INPUT_DECAY_IN]*ms

# from hereon, all code, until the runsimulation command is compiled to c++, variable values (parameter arrays etc) currently
# in the python environment are also made available to the standalone executable
if standalone:
    set_device('cpp_standalone',directory=simulation_output.get_output_dir_path()+'output')
    prefs.codegen.cpp.headers += ['"run.h"'] # This is necessary to use brian_end()

#assign some of the parameters here in the local scope to be used in the model
defaultclock.dt = 1*ms
tau_s=1*ms
tau_clock_mult = defaultclock.dt/ms

eqs = '''
dIe/dt = Ae/tau_s + (-Ie/tauI_e)  : 1
dIi/dt = Ai/tau_s + (-Ii/tauI_i)  : 1
dp/dt = (Ii+Ie)/tau_s + (background_p - p) /taup : 1
dAe/dt = -Ae/tau_s  : 1
dAi/dt = -Ai/tau_s  : 1
background_p : 1
reset_p : 1
taup : second
'''

########################################## REALTIME VISUALIZATION OF NETWORK ACTIVITY
#to enable set the 'Show sim GUI' flag in the simulation parameter file to true
#note ... it slows down activity very much
plotting_parameters = p.get_plotting_parameters()
if show_gui:
    from cros_model.graphics.live_plotting import Live_plotting_window,Plot_data

    plot_data = Plot_data(plotting_parameters,parameters,types,excitatory_indexes_sorted,inhibitory_indexes_sorted,activity_variable_name='p',activity_variable_unit=1)

    #TODO: this for one works only when replotting at every 1ms (and only when using 1ms as simulation dt),
    #for larger intervals I would have to make parallel network operation that sets
    #last spikes at every timestep, and computes the last spikes
    @network_operation(dt=1*ms,when='end')
    def data_update(clocktime):
        if clocktime>1*ms:
            exc_p = neurongroup_exc.p
            inh_p = neurongroup_inh.p

            all_p = np.zeros(num_excitatory+num_inhibitory)
            all_p[excitatory_indexes_sorted]=exc_p
            all_p[inhibitory_indexes_sorted]=inh_p

            input_data = {'clocktime':clocktime,
                          'num_excitatory':num_excitatory,
                          'num_inhibitory':num_inhibitory,
                          'exc_lastspikes':neurongroup_exc.lastspike,
                          'inh_lastspikes':neurongroup_inh.lastspike,
                          'population_monitor_exc':population_monitor_exc,
                          'population_monitor_inh':population_monitor_inh,'current_grid_activation':all_p}
            plot_data.calculate_new_plot_data(input_data)

        speed = plot_data.get_speed()
        if speed=="min":
            sleep(0.50)
        if speed=="slow":
            sleep(0.25)
        elif speed=="medium":
            sleep(0.01)
        elif speed=="high":
            sleep(0.003)

    window = Live_plotting_window(plot_data,)

    net=Network(data_update)
else:
    net=Network()

neurongroup_exc = NeuronGroup(num_excitatory, eqs, threshold='rand()<=p', reset='p = reset_p',refractory=0*ms,method='euler')
neurongroup_exc.p=background_p_exc
neurongroup_exc.background_p= background_p_exc
neurongroup_exc.reset_p=P_reset_exc
neurongroup_exc.taup=taup_exc
net.add(neurongroup_exc)

neurongroup_inh = NeuronGroup(num_inhibitory, eqs, threshold='rand()<=p', reset='p = reset_p',refractory=0*ms,method='euler')
neurongroup_inh.p=background_p_inh
neurongroup_inh.background_p= background_p_inh
neurongroup_inh.reset_p=P_reset_inh
neurongroup_inh.taup=taup_inh
net.add(neurongroup_inh)

if parameters[p.EXTERNAL_INPUT]==True and parameters[p.STIMULUS_SIZE]>0:
    #set up network external input - this assumes default clock at millisecond
    external_input_times=np.zeros(parameters[p.SIMULATION_LENGTH]-2)
    
    #this is assumes clock time of 1ms
    for i in xrange(len(external_input_times)):
        external_input_times[i]=750+rand()*500+1000*i
    
    neurongroup_stim = SpikeGeneratorGroup(stimulus_size, \
                                           np.repeat(np.arange(stimulus_size),len(external_input_times)), \
                                           np.tile(external_input_times,stimulus_size)*ms)
    net.add(neurongroup_stim)

interval_weight_update = 5

########################################## SYNAPSE MODEL

ee_gain = parameters[p.EE_GAIN]
ei_gain = parameters[p.EI_GAIN]
ie_gain = parameters[p.IE_GAIN]
ii_gain = parameters[p.II_GAIN]

neuron_model = 'w :1'
pre_model_ee = 'Ae+=ee_gain*w'
pre_model_ei = 'Ae+=ei_gain*w'
pre_model_ie = 'Ai+=ie_gain*w'
pre_model_ii = 'Ai+=ii_gain*w'

#no gain applied to input layer
pre_model_external_input = 'Ae+=w'
Cee = Synapses(neurongroup_exc, neurongroup_exc,neuron_model, pre=pre_model_ee,method='euler')
Cei = Synapses(neurongroup_exc, neurongroup_inh,neuron_model, pre=pre_model_ei,method='euler')
Cie = Synapses(neurongroup_inh, neurongroup_exc,neuron_model, pre=pre_model_ie,method='euler')
Cii = Synapses(neurongroup_inh, neurongroup_inh,neuron_model, pre=pre_model_ii,method='euler')

if parameters[p.EXTERNAL_INPUT]==True and parameters[p.STIMULUS_SIZE]>0:  
    Cse = Synapses(neurongroup_stim, neurongroup_exc,neuron_model, pre=pre_model_external_input,method='euler')

Cee.connect(i=ee_from_idx,j=ee_to_idx)
Cei.connect(i=ei_from_idx,j=ei_to_idx)
Cie.connect(i=ie_from_idx,j=ie_to_idx)
Cii.connect(i=ii_from_idx,j=ii_to_idx)

Cee.w = wEE
Cei.w = wEI
Cie.w = wIE
Cii.w = wII

net.add(Cee)
net.add(Cei)
net.add(Cie)
net.add(Cii)

########################################## ADD ACTIVITY MONITORS

if parameters[p.EXTERNAL_INPUT]==True and parameters[p.STIMULUS_SIZE]>0:
    Cse.connect(i=stimulus_input_from_idx,j=stimulus_input_to_idx)
    Cse.w = w_external_exc
    net.add(Cse)

# add activity monitors
population_monitor_exc = PopulationRateMonitor(neurongroup_exc)
net.add(population_monitor_exc)

population_monitor_inh = PopulationRateMonitor(neurongroup_inh)
net.add(population_monitor_inh)

if parameters[p.OUTPUT_ACTIVITY]:
    s_mon_exc = SpikeMonitor(neurongroup_exc)
    s_mon_inh = SpikeMonitor(neurongroup_inh)
    net.add(s_mon_exc)
    net.add(s_mon_inh)

########################################## RUN SIMULATION

simduration = parameters[p.SIMULATION_LENGTH] * second
numcycles = int(simduration / ms)

net.run(simduration,report='stdout')

########################################## PUT TOGETHER ALL MONITORED VARIABLES FOR OUTPUT AND ANALYSIS

simulation_variables = {'numcycles': numcycles, 'real_path': simulation_output.get_output_dir_path(), 'Cee': Cee,
                        'Cei': Cei, 'Cie': Cie, 'Cii': Cii, 'net':net,'types':types,
                        'num_excitatory':num_excitatory,'num_inhibitory':num_inhibitory,
                        'population_monitor_exc':population_monitor_exc,'population_monitor_inh':population_monitor_inh,
                        'excitatory_indexes_sorted':excitatory_indexes_sorted,'inhibitory_indexes_sorted':inhibitory_indexes_sorted}

if parameters[p.OUTPUT_ACTIVITY]:
    simulation_variables['s_mon_exc']=s_mon_exc
    simulation_variables['s_mon_inh']=s_mon_inh

if parameters[p.EXTERNAL_INPUT]==True and parameters[p.STIMULUS_SIZE]>0:
    simulation_variables['stimulus_size']=stimulus_size
    simulation_variables['Cse']=Cse
    simulation_variables['external_input_times']=external_input_times
else:
    simulation_variables['stimulus_size']=0
    simulation_variables['Cse']=[]
    simulation_variables['external_input_times']=[]

########################################## OUTPUT, ANALYSE, PLOT   
simulation_output.output_simulation(simulation_variables)

matlab_analysis.run(simulation_output.get_output_dir_path())

########################################## CLEANUP
#delete generated code directory

if parameters[p.STANDALONE_MODE]:
    generated_code_directory = simulation_output.get_output_dir_path()+'output'
    if os.path.isdir(generated_code_directory):
        shutil.rmtree(generated_code_directory)